import click

from baseten.common import api
from baseten.common.model_deployer import pull_model


@click.group()
def models():
    """Manage your models hosted on Baseten"""
    pass


@models.command()
def list():
    """Lists all user models"""
    pretty_print_models(api.models()["models"])


@models.command()
@click.argument("model_name")
def get(model_name):
    """Gets model information given name"""
    pretty_print_models([api.get_model(model_name)["model_version"]["oracle"]])


@models.command()
@click.option("--model_version_id", prompt="Model Version Id")
@click.option("--directory", prompt="Target Directory")
def pull(model_version_id, directory=None):
    """Pulls models given model_version id"""
    pull_model(model_version_id, directory)


def pretty_print_models(models):
    for model in models:
        semver_to_version_id = []
        for model_version in model["versions"]:
            semver_to_version_id.append({f"Version {model_version['semver']}": model_version["id"]})
        click.echo(f"Name: {model['name']} | Versions: {semver_to_version_id}")



from typing import Any

from baseten.baseten_deployed_model import BasetenDeployedModel
from baseten.common import api, settings
from baseten.models.util import get_or_create_pretrained_model, requests_error_handling


class FlanT5:
    """
    Flan-T5 is a instruction-tuned text-to-text generation model from Google. It is built atop the generic T5 model.

    **Examples:**

    Invoking default Flan T5

    ```python
    from baseten.models import FlanT5
    model = FlanT5()
    model("The quick brown fox jumps over the lazy dog.")
    ```

    **Setting bad words**

    For production apps, you may find it useful to set a list of "bad words" which tell the model
    what words it cannot use during generation. Blueprint supports this by letting you pass that
    list during instantiation of the model.

    ```python
    model = FlanT5()
    model("The quick brown fox jumps over the lazy dog.", bad_words=["dog"])
    ```

    ```python
    model = FlanT5()
    model.bad_words = ["dog"]
    model("The quick brown fox jumps over the lazy dog.")
    ```

    ```python
    model = FlanT5()
    model.bad_words = ["dog"]
    model("The quick brown fox jumps over the lazy dog.", num_beams=4)
    ```
    """

    def __init__(self):
        """
        Attributes:
            bad_words (list): List of words to avoid in the output.
        """
        self.bad_words = []
        self._model = self._set_user_model()

    def _set_user_model(self) -> BasetenDeployedModel:
        """Creates internal BasetenDeployedModel object that points to users
        deployed Flan-T5 model. If the user does not have a deployed model,
        we will create one for them.
        """
        model_version = get_or_create_pretrained_model("Flan-T5 XL")
        return BasetenDeployedModel(
            model_id=model_version["model_id"],
            truss_spec_version=model_version["truss_spec_version"],
        )

    def __setattr__(self, __name: str, __value: Any) -> None:
        """Validates attributes before setting them."""
        if __name == "bad_words":
            if not isinstance(__value, list):
                raise ValueError("bad_words must be a list")
        super().__setattr__(__name, __value)

    def __call__(self, prompt, seed=None, **kwargs) -> str:
        """Generate text from a prompt. Supports all parameters of the underlying model
        from Huggingface library.

        Args:
            prompt (str): The prompt to generate text from.
            seed (int): The random seed to use for reproducibility. Optional.
        Attributes:
            num_beams (int): Number of beams for beam search. 1 means no beam search. Optional. Default to 1.
            num_return_sequences (int): The number of independently computed returned sequences for
                each element in the batch. Default to 1.
            max_new_tokens (int): The maximum number of tokens that can be generated. Default to 50.
            temperature (float): The value used to module the next token probabilities. Must be
                strictly positive. Default to 1.0.
            top_k (int): The number of highest probability vocabulary tokens to keep for top-k-filtering.
                Between 1 and infinity. Default to 50.
            top_p (float): If set to float < 1, only the most probable tokens with probabilities that
                add up to top_p or higher are kept for generation. Default to 1.0.
            repetition_penalty (float): The parameter for repetition penalty. Between 1.0 and infinity.
                1.0 means no penalty. Default to 1.0.
            length_penalty (float): Exponential penalty to the sequence length. Default to 1.0.
            no_repeat_ngram_size (int): If set to int > 0, all ngrams of that size can only occur
                once. Default to 0.
            num_return_sequences (int): The number of independently computed returned sequences
                for each element in the batch. Default to 1.
            do_sample (bool): If set to False greedy decoding is used. Otherwise sampling is used.
                Defaults to True.
            early_stopping (bool): Whether to stop the beam search when at least num_beams sentences
                are finished per batch or not. Default to False.
            use_cache (bool): Whether or not the model should use the past last key/values attentions
                (if applicable to the model) to speed up decoding. Default to True.
            decoder_start_token_id (int): If an encoder-decoder model starts decoding with a different
                token than BOS, the id of that token. Default to None.
            pad_token_id (int): The id of the padding token. Default to None.
            eos_token_id (int): The id of the end of sequence token. Default to None.
            forced_bos_token_id (int): The id of the token to force as the first generated
                token after the BOS token. Default to None.
            forced_eos_token_id (int): The id of the token to force as the last generated
                token when max_length is reached. Default to None.
            remove_invalid_values (bool): Whether or not to remove possible `nan` and `inf`
                outputs of the model to prevent the generation method to crash. Default to False.

        Returns:
            generated_text (str): The generated text.
        """
        request_body = {"prompt": prompt, **kwargs}

        if self.bad_words:
            request_body["bad_words"] = self.bad_words
        if seed:
            request_body["seed"] = seed

        with requests_error_handling():
            server_response = self._model.predict(request_body)

        if server_response["status"] == "error":
            raise ValueError(server_response["message"])
        return server_response["data"]

    @staticmethod
    def url():
        """
        Use this static method to get a URL to the Stable Diffusion model page
        in your Blueprint project, which contains information about the model.

        **Example:**

        ```python
        from baseten.models import FlanT5

        FlanT5.url()
        ```
        """
        try:
            blueprint_project_id = api.get_blueprint_projects()[0]["id"]
            return (
                f"{settings.get_server_url()}/blueprint/projects/"
                f"{blueprint_project_id}?est=community-model--flan"
            )
        except IndexError:
            raise ValueError("User does not have any blueprint projects.")



import json
import os
from json import JSONDecodeError
from typing import Optional

import click

from baseten.training import (
    Dataset,
    DatasetIdentifier,
    DreamboothConfig,
    FinetuningRun,
    LocalPath,
    PublicUrl,
)


def _finetune_interactive(finetuning_type: str):
    raise NotImplementedError


def _validate_input_options(input_: Optional[str], input_file: Optional[str]) -> bool:

    if input_ and input_file:
        click.echo("Please provide exactly one of input|input_file")
        return False

    if input_ is None and input_file is None:
        click.echo("Please provide exactly one of input|input_file")
        return False

    if input_file is not None and (not os.path.exists(input_file) or os.path.isdir(input_file)):
        click.echo(f"{input_file} is not a valid file.")
        return False

    return True


def _parse_input_data(params: dict) -> Optional[DatasetIdentifier]:
    if "input_dataset" in params:
        raw_input_data = params["input_dataset"]
    else:
        return None

    if "public_url" in raw_input_data:
        return PublicUrl(raw_input_data["public_url"])

    if "baseten_dataset_id" in raw_input_data:
        return Dataset(raw_input_data["baseten_dataset_id"])

    if "local_path" in raw_input_data:
        if "dataset_name" in raw_input_data:
            return LocalPath(raw_input_data["local_path"], raw_input_data["dataset_name"])
        else:
            click.echo("If using local_path, please also provide a `dataset_name`.")
            return None

    return None


def _finetune_dreambooth(trained_model_name: str, params: dict) -> Optional[FinetuningRun]:
    parsed_input_dataset = _parse_input_data(params)

    if parsed_input_dataset:
        params.update({"input_dataset": parsed_input_dataset})
        config = DreamboothConfig(**params)
        return FinetuningRun.create(trained_model_name, config)
    else:
        click.echo("No valid `input_dataset` provided.")
        return None


@click.group(name="finetuning")
def finetuning_cli():
    """Run custom Finetuning jobs with Blueprint"""
    pass


@finetuning_cli.command()
@click.option(
    "--interactive/--no-interactive",
    default=True,
    help="Deploy interactively using a series of guided prompts.",
)
@click.option(
    "--finetuning-type",
    type=click.Choice(["DREAMBOOTH", "CLASSIC_STABLE_DIFFUSION"]),
    required=True,
    help="Type of Finetuning run to create.",
)
@click.option("--name", required=True, help="Name for this Finetuning run.")
@click.option("--input-file", required=False, help="File pointing to the configuration.")
# TODO(sid): What is this?
@click.option("--input", "input_", required=False)
def create(interactive: bool, finetuning_type: str, name: str, input_file: str, input_: str):
    """Create and trigger new fine-tunng run

    See
    [FinetuningRun.create][baseten.training.finetuning.FinetuningRun.create]
    for more details.
    """
    if interactive:
        click.echo("Starting interactive finetuning flow.")
        _finetune_interactive(finetuning_type)
    else:
        if _validate_input_options(input_, input_file):
            # TODO: Take input or input file, parse it into a dict. Handle the input_data piece, and
            # construct a Dreambooth config object.
            if input_file:
                input_ = open(input_file, "r").read()

            try:
                parsed_json = json.loads(input_)
            except JSONDecodeError as error:
                click.echo("Could not parse JSON.")
                click.echo(str(error))
                return

            if finetuning_type == "DREAMBOOTH":
                training_run = _finetune_dreambooth(name, parsed_json)
                if training_run:
                    click.echo(f"Started training run: {training_run.id}")
                    click.echo(f"Access the training logs at {training_run.blueprint_url}")
                    click.echo(f"or use `blueprint finetunings logs --id {training_run.id}")


@finetuning_cli.command()
def list():
    """List all finetuning runs.

    See
    [FinetuningRun.list][baseten.training.finetuning.FinetuningRun.list]
    for more details.
    """
    runs = FinetuningRun.list()
    if len(runs) == 0:
        click.echo("No runs found. Kick off a run using `baseten finetuning create`.")
        return
    click.echo(f"Found {len(runs)} runs")
    for run in runs:
        click.echo(4 * " " + f"id: {run.id}, name: {run.trained_model_name}")


@finetuning_cli.command()
@click.option("--run-id", type=str, required=True, help="ID for Blueprint Finetuning Run.")
@click.option(
    "--idle-time-minutes",
    type=int,
    default=30,
    help="Number of minutes before server scales down to save costs.",
)
def deploy(run_id: str, idle_time_minutes: int):
    """Deploy model from finetuning run


    See
    [FinetuningRun.deploy][baseten.training.finetuning.FinetuningRun.deploy]
    for more details.
    """
    deployed_model = FinetuningRun(run_id).deploy(idle_time_minutes)
    click.echo(f"Access the deployed model at {deployed_model.blueprint_url}")


@finetuning_cli.command()
@click.option("--run-id", type=str, required=True, help="ID for Blueprint Finetuning Run.")
def cancel(run_id: str):
    """Cancel Finetuning run

    See
    [FinetuningRun.cancel][baseten.training.finetuning.FinetuningRun.cancel]
    for more details.
    """
    cancelled = FinetuningRun(run_id).cancel()
    if cancelled:
        click.echo(f"Successfully cancelled finetuning run {run_id}.")
    else:
        click.echo(f"Failed to cancel finetuning run {run_id}.")


@finetuning_cli.command()
@click.option("--run-id", type=str, required=True, help="ID for Blueprint Finetuning Run.")
def logs(run_id: str):
    """Stream logs for Finetuning run

    See
    [FinetuningRun.stream_logs][baseten.training.finetuning.FinetuningRun.stream_logs]
    for more details."""
    FinetuningRun(run_id).stream_logs()



"""Top level training package for baseten.

Exposes relavant classes from submodules for DX purposes and discoverability
"""
from baseten.training.datasets import Dataset, DatasetIdentifier, LocalPath, PublicUrl
from baseten.training.finetuning import (
    DreamboothConfig,
    FinetuningConfig,
    FinetuningRun,
    FullStableDiffusionConfig,
)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from baseten.common.api import get_training_dataset
from baseten.common.files import DatasetTrainingType, upload_dataset


class DatasetIdentifier(ABC):
    """Base class for all the possible ways to indentify a dataset.
    You must provide a DatasetIdentifier to a FinetuneConfig to supply input data for the FinetuningRun.
    """

    @abstractmethod
    def resolve_dataset_url(self, training_type: DatasetTrainingType) -> str:
        """
        Method that returns a URL from which the training job can download
        a training dataset.
        """


class Dataset(DatasetIdentifier):
    """
    A Dataset hosted on Baseten

    **Example:**

    ```python
    from baseten.training import Dataset
    dataset = Dataset("DATASET_ID")
    ```
    """

    def __init__(self, dataset_id: str):
        """
        Args:
            dataset_id: The ID of the dataset hosted on Baseten
        """
        self.dataset_id = dataset_id

    def resolve_dataset_url(self, training_type: DatasetTrainingType) -> str:
        training_dataset = get_training_dataset(dataset_id=self.dataset_id)
        return training_dataset["presigned_s3_get_url"]


class LocalPath(DatasetIdentifier):
    """
    A local dataset to be uploaded

    **Example:**

    ```python
    from baseten.training import LocalPath
    dataset = LocalPath("./my-dataset")
    ```

    """

    def __init__(self, path: str, dataset_name: Optional[str] = None):
        """
        Args:
            path: The absolute or relative path to the dataset directory on your local machine
            dataset_name: The name to assign the dataset once it is uploaded to Baseten.
                If none, the uploaded directory's name will be used instead.
        """
        self.path: Path = Path(path)
        self.dataset_name: Optional[str] = dataset_name

    def resolve_dataset_url(self, training_type: DatasetTrainingType) -> str:
        dataset_id = upload_dataset(
            dir=self.path, training_type=training_type, name=self.dataset_name
        )
        training_dataset = get_training_dataset(dataset_id=dataset_id)

        return training_dataset["presigned_s3_get_url"]


class PublicUrl(DatasetIdentifier):
    """
    A dataset hosted at a publicly accessible url

    **Example:**

    ```python
    from baseten.training import PublicUrl
    dataset = PublicUrl("https://cdn.baseten.co/docs/production/DreamboothSampleDataset.zip")
    ```
    """

    def __init__(self, url: str):
        """
        Args:
            url: The URL of a zip file to use as a dataset. This URL must be publicly accessible (unauthenticated)
        """
        self.url = url

    def resolve_dataset_url(self, training_type: DatasetTrainingType) -> str:
        return self.url



snippet

from baseten.training import finetune
from baseten.training import Dataset, PublicUrl
from baseten.training import DreamboothConfig

input_data = Dataset("ZWB5Lqk")
# OR input_data = PublicUrl("https://path-to-dataset.com")
config = DreamboothConfig(
    instance_prompt="photo of sks",
    input_dataset=input_data,
    wandb_api_key="83b673c1371752f6177aafc403be4663970505c1",
)

fr = finetune("sid-photo-model-12", config)



import logging
import tarfile
import tempfile
from typing import List, Optional

import requests
from colorama import Fore, Style
from tqdm import tqdm

from baseten.baseten_deployed_model import BasetenDeployedModel
from baseten.common import api
from baseten.common.core import raises_api_error

logger = logging.getLogger(__name__)


class BasetenArtifact:
    """An artifact stored on the Baseten backend"""

    def __init__(
        self,
        artifact_id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        deleted: bool = False,
    ):
        if not artifact_id:
            raise ValueError("Must provide the artifact's id")

        self._id = artifact_id
        self._name = name
        self._description = description
        self._deleted = deleted

    def __repr__(self):
        attr_info = [f"id={self._id}"]
        if self._deleted:
            attr_info.append("DELETED ON SERVER")
        if self._name:
            attr_info.append(f"name={self._name}")
        if self._description:
            attr_info.append(f"description={self._description}")

        info_str = "\n  ".join(attr_info)

        return f"BasetenArtifact<\n  {info_str}\n>"

    @raises_api_error
    def create_link(self, model_id: Optional[str] = None, model_version_id: Optional[str] = None):
        if not model_id and not model_version_id:
            raise ValueError("Either model_id or model_version_id must be provided.")
        if model_id and model_version_id:
            raise ValueError("Must provide either model_id or model_version_id; not both.")
        resp = api.create_artifact_link(
            self._id, model_id=model_id, model_version_id=model_version_id
        )
        logger.info("⛓ Successfully linked artifact ⛓")
        return resp

    @raises_api_error
    def links(self):
        links = api.artifact_links(self._id)
        return [
            BasetenDeployedModel(model_version_id=model_version_id)
            for model_version_id in links["model_version_ids"]
        ] + [BasetenDeployedModel(model_id=model_id) for model_id in links["model_ids"]]

    @raises_api_error
    def url(self) -> str:
        # todo(alex): worth it to introduce a library to cache a TTL here?
        return api.artifact_url(self._id)

    @raises_api_error
    def download(self, target_directory: str):
        artifact_response = requests.get(self.url(), stream=True)
        logger.info("📁 Downloading artifact archive from Baseten 📁")
        block_size = 1024  # 1 Kibibyte
        total_size_in_bytes = int(artifact_response.headers.get("content-length", 0))
        progress_bar = tqdm(
            total=total_size_in_bytes,
            unit="iB",
            unit_scale=True,
            bar_format="Download Progress: "
            "{percentage:3.0f}%% |%s{bar:100}%s| {n_fmt}/{total_fmt}" % (Fore.BLUE, Fore.RESET),
        )

        temp_tgz = tempfile.NamedTemporaryFile(mode="w+b")
        for content in artifact_response.iter_content(block_size):
            temp_tgz.write(content)
            progress_bar.update(len(content))
        temp_tgz.file.seek(0)
        progress_bar.close()

        logger.info("🔮 Download successful!🔮")

        with tarfile.open(temp_tgz.name, "r") as tar:
            tar.extractall(target_directory)
        logger.info(f"📁 Successfully extracted artifact to {target_directory} 📁")

    @raises_api_error
    def update(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        list_of_files: Optional[List[str]] = None,
    ):
        s3_key = None
        artifact_name = name or self._name
        artifact_description = description or self._description
        if list_of_files:
            s3_key = api.serialize_artifact_to_s3(artifact_name, list_of_files)
        update_resp = api.update_artifact(self._id, artifact_name, artifact_description, s3_key)
        self._id = update_resp["id"]
        self._name = update_resp["name"]
        self._description = update_resp["description"]

    @raises_api_error
    def delete(
        self,
    ):
        api.delete_artifact(self._id)
        self._deleted = True


@raises_api_error
def create_artifact(
    name: str,
    list_of_files: List[str],
    description: Optional[str] = None,
    model_id: Optional[str] = None,
    model_version_id: Optional[str] = None,
):
    """Create an artifact in the Baseten backend to track

    Args:
        name (str): The name of the artifact to be created.
        list_of_files (List[str]): A list of files to be uploaded.
        description (str, optional): An optional description of the artifact
        model_id (str, optional): A model_id to associate with the artifact
        model_version_id (str, optional): A model_version_id to associate with the artifact
    Returns:
        BasetenArtifact (BasetenArtifact): An object representing the newly created artifact

    Raises:
        ApiError: If there was an error communicating with the server.
    """
    s3_key = api.serialize_artifact_to_s3(name, list_of_files)
    artifact_resp = api.create_artifact(
        s3_key, name, description=description, model_id=model_id, model_version_id=model_version_id
    )
    logger.info(
        f'Successfully stored {Fore.BLUE}{name}{Style.RESET_ALL} artifact with id {artifact_resp["artifact_id"]}.'
    )
    return BasetenArtifact(**artifact_resp)
